/*
 * Aphelion
 * Copyright (c) 2013  Joris van der Wel
 * 
 * This file is part of Aphelion
 * 
 * Aphelion is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, version 3 of the License.
 * 
 * Aphelion is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with Aphelion.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * In addition, the following supplemental terms apply, based on section 7 of
 * the GNU Affero General Public License (version 3):
 * a) Preservation of all legal notices and author attributions
 * b) Prohibition of misrepresentation of the origin of this material, and
 * modified versions are required to be marked in reasonable ways as
 * different from the original version (for example by appending a copyright notice).
 * 
 * Linking this library statically or dynamically with other modules is making a
 * combined work based on this library. Thus, the terms and conditions of the
 * GNU Affero General Public License cover the whole combination.
 * 
 * As a special exception, the copyright holders of this library give you 
 * permission to link this library with independent modules to produce an 
 * executable, regardless of the license terms of these independent modules,
 * and to copy and distribute the resulting executable under terms of your 
 * choice, provided that you also meet, for each linked independent module,
 * the terms and conditions of the license of that module. An independent
 * module is a module which is not derived from or based on this library.
 */

package aphelion.shared.physics.operations;


import aphelion.shared.physics.PhysicsEnvironment;
import aphelion.shared.physics.State;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author Joris
 */
public class LoadConfig extends Operation
{
        private static final Logger log = Logger.getLogger("aphelion.shared.physics");

        public String fileIdentifier;
        public List<Object> yamlDocuments;
        
        private final boolean[] executed = new boolean[PhysicsEnvironment.MAX_TRAILING_STATES];
        
        public LoadConfig()
        {
                super(false, PRIORITY.LOAD_CONFIG);
        }
        
        @Override
        public boolean execute(State state, long ticks_late)
        {
                
                // Note: Side effects of config change is not checked for consistency explicitely.
                // (Only effects on position and energy are checked at the moment).
                // Preferably the server should send these load config operations with a future tick 
                // so that any inconsistency is minimized.
                // For example it is legal to send this operation with a tick that is 10 seconds from now
                
                // Thanks to periodic Actor Syncing any inconsistency caused by late LoadConfig will be resolved eventually.

                
                
                if (ticks_late == 0)
                {
                        log.log(
                                Level.INFO, 
                                "Loading {0} yaml documents with file identifier {1} in state {2} ...", 
                                new Object[] {yamlDocuments.size(), fileIdentifier, state.id});
                }
                else
                {
                        log.log(
                                Level.WARNING, 
                                "Loading {0} yaml documents with file identifier {1} in state {2}  (LATE {3}!)  ...", 
                                new Object[] {yamlDocuments.size(), fileIdentifier, state.id, ticks_late});
                }
                
                state.config.addFromYaml(yamlDocuments, fileIdentifier);
                state.config.applyChanges();
                
                if (this.tick > state.config_lastModification)
                {
                        state.config_lastModification = this.tick;
                }
                
                
                executed[state.id] = true;
                return true;
        }

        @Override
        public boolean isConsistent(State older, State newer)
        {
                if (this.tick > older.tick_now)
                {
                        return true; // This operation has not had the chance to execute on both states (yet)
                }
                
                return executed[older.id] == executed[newer.id];
        }

        @Override
        public void resetExecutionHistory(State state, State resetTo)
        {
                executed[state.id] = executed[resetTo.id];
        }
}
